-- attrKey.ms - By Michael B. Comet - comet@comet-cartoons.com
--				http://www.comet-cartoons.com/
--
--	Copyright 2001 Michael B. Comet  All Rights Reserved
--	The script is provided AS IS, there are no warranties about
--  it's usability, safety, or bugs.  Use at your own risk!
--
--
-- This script allows you to easily view, and animate attributes
-- on any number of selected objects, as well as animate Position,
-- Rotation, or scale for the objects.
--
-- Only float, integer, booleans and color type attributes are supported,
-- since they are the most common animateable type attributes.
--
-- Make sure you have MAX's Animate Mode ON when animating!
--
-- Note: this assumes Boolean attrs use On/Off controllers, and so
--  they won't get keyed, unless their value is supposed to change.
--  Normally, this should act like you want it.
--
-- Note: Objects with spaces in the name, will NOT work.
-- Make sure things like end effectors are converted to use _ or
-- something else, and not spaces.
--
--
-- Rev 1.01, fixed up the create_attr_ui() function so it checks MODIFIERS for
--		custom attributes.  So Modifier Based CA's are listed too.
--		Also, keying items or attributes is now Undoable.
--
-- Rev 1.02 -  March 1, 2001, fixed up create_attr_ui() function so it also
-- 		checks the MATERIALS for custom attributes.  Now ALL custom attributes
--		for an object should appear on the list.
--
--
-- Rev 1.03 - March 3, 2001, now ints and floats use the correct attribute range
--		for the original added attr.  Yay.
--
-- Rev 1.10 - August 8, 2001: Now has disable and enable Scene Redraw calls for
--		faster speed when keying.
--		Also added new feature for being able to select, and key PRS/attr data
--			for individual objects only in attrKey.  The select is mostly useful
--			if the auto-refresh/update is disabled..so it becomes a kind of U.I.
--
-- Rev 1.11 - August 8, 2001:
--		Fixed bug where range wasn't working in MAX 4.2 due to missing "'" around
--			name of the attr.
--		Fixed bug where the new per Obj keying stuff broke non float type attrs.
--
-- Rev 1.12  - August 8, 2001:
--		Fixed redraw bug where screen redraw would stop if object with no attrs
--			was keyed
--		Objects with spaces in name now supported.
--
-- Rev 1.13  - September 30, 2001:
--		Fixed range function break when a custom defdata didn't exist.
--
-- Rev 1.14  - October 3, 2001:
--		Removed printing of debug notes in get range.
--
-- Rev 1.20 - October 6, 2001:
--		Removed "Custom_Attributes" since not needed when accessing stuff and if
--		there were multiple instances of them or attr structs named different, this wouldn't work
--
-- Rev 1.21 - October 10, 2001:
--		Fixed NASTY bug where "undo" on a selection would cause MAX to crash.
--		This was due to call gc() garbage collect inside of the selChanged callback.
--		Also made undo calls have a nice string.
--
-- Rev 1.22 - October 13, 2001:
--		Added a try/catch to the modifier CA check.  This fixes errors when run
--		on objects that had some "BETA" modifier on it that aren't returning data
--		the same way or something.  Probably safer to do that anyhow.
--
-- Rev 1.23 - October 14, 2001:
--		Added POSE saving and loading and copy paste!!! Yay!
--		Added support for SLIDER Manipulator objects to show value in attrKey
--
-- Rev 1.24 - October 16, 2001:
--		Fixed fwd reference of save/load POSE functions by declaring global on top.
--		Made POSE restore position last to get rid of rotation causing position shifts.
--		Fixed rotation pose saving to use Euler Angle output for easier editing by hand
--			of the pose text file.
--		Added Wait cursor to pose saving and loading.
--
-- Rev 1.25 - October 17, 2001:
--		Changed stuff to use parent coordsys for true fixing of pose save/load issues.
--
-- Rev 1.30 - March 2, 2002:
--		Added Pos and Rot and Scale  spinners for Position_XYZ and Euler_XYZ, ScaleXYZ to the UI.
--
-- Rev 1.31 - March 3, 2002:
--		Added support for ZEROing out PRS and Attr data on All or Specific objects.
--
-- Rev 1.40 - March  9, 2002
--		Added new POSE options for loading/pasting only the PRS or only the Attr data.
--		Also added new Blending function to allow the loaded/pasted pose to be blended with
--			the state the objects were in when the load/paste button was hit.
--		Made zero attr stuff redrawViews to make sure things show up right.
--
-- Rev 2.00 - April 27, 2002
--		This version has been recreated to use StringStreams for most stuff.  This
--		makes attrKey MUCH more efficient and less likely to crash.  Thanks to Larry
--		Minton at Discreet for this fix...as he did most of the work actually.
--		I have also fixed up this version so that CA's with spaces in the name now work ok.
--
-- Rev 2.01 - October 21, 2002
--		Added code to handle setKeyBrackets on spinners, so that they highlight red when
--		keyed.  This only works if you a) update time, b) hit the "K" button or c) load a pose
--		d) Zero or Key all Attrs.
--		If you adjust a spinner it won't highlight red or not based on that since there is
--		no handle callback and it is instead done via controller hookups.
--
-- Rev 2.02 - October 26, 2002
--		Removed keyBracket code since r5.1 will do that automatically.  And it's faster and
--		more stable.
--		Added new option for Slider vs. Spinner ui display for float/ints.  Also fixed
--		drawing of all items so that things line up nicer.
--

-- Globals
aKtitlever = "2.02";
global attrKeyRunning;
global aK_objs = #();			-- object name list
global aK_attr = #();			-- attr name list
global aK_cntrl = #();			-- attr controller list
global aK_attrUI = #();			-- ui item list
global aK_type = #();			-- type of attr list
global aK_start = #();			-- start attr idx for given obj
global aK_end = #();			-- end attr array idx for given obj

	-- Stuff for load/paste Pose Blending...
global aK_poseCntrl = #();		-- The control+value string that will set the pose
global aK_poseStart = #();		-- The initial value
global aK_poseEnd = #();		-- the end loaded value
global aK_poseType= #();		-- Type of value


global attrKeyRF;
global attrRoll;
global savePose;
global loadPose;
global update_attr_ui;

global int aKShowPRS = 1;		-- should we show PRS data?
global int aKloadPRS = 1;		-- should we load PRS when we do a load/paste pose?
global int aKloadAtt = 1;		-- should we load Attrs when we do a load/paste pose?

global int aKuseSliders = 0;	-- should we use sliders or spinners?


-- ---------------------- Nice StringStream & String Functions -----------------


fn executeString s = (if classof s == stringstream then seek s 0; execute s)
fn appendString s add = (format "%" add to:s)

fn space2Under s =
(
	n = s.count;
	retstr =  "";

	for i in 1 to n do
	    (
		if (s[i] == " ") then
		    retstr += "_";
		else
		    retstr += s[i];
		)

	return retstr;
)


-- -------------------------- Pose Functions ---------------------


/*
 * isSelected() - Returns true if the passed in object is selected
 *						  false otherwise.
 */
global isSelected;	-- Make this a GLOBAL function callable from now on.
fn isSelected obj =
(
    sel = getCurrentSelection();
	idx = findItem sel obj;
	if (idx > 0) then
		return true;
	else
		return false;
)

/*
 * aK_copyPose() - Wrapper call for copying a pose
 */
fn aK_copyPose =
(
	thePath = GetDir #scripts;
	r = savePose (thePath + "\\attrKeyCopyPastePOSE.txt")
	if (r == true) then
	    format "-- POSE Copied.\n";
)

/*
 * aK_pastePose() - Wrapper call for pasting a pose
 */
fn aK_pastePose =
(
	thePath = GetDir #scripts;
	r = loadPose (thePath + "\\attrKeyCopyPastePOSE.txt")
	if (r == true) then
	    format "-- POSE Pasted.\n";
)

/*
 * aK_savePose() - Wrapper call for saving a pose
 */
fn aK_savePose =
(
	thePath = GetDir #scripts;
    filename = getSaveFileName caption:"Save POSE" filename:(thePath+"\\") types:"PoseTXT (*.txt)|*.txt|All (*.*)|*.*";
	if (filename == undefined) then		-- did the user cancel?
		return;

	format "-- savePOSE: %\n" filename;

	r = savePose filename;
	if (r == true) then
   		format "-- POSE % Saved.\n" (getFileNameFile(filename));
)

/*
 * aK_loadPose() - Wrapper call for loading a pose
 */
fn aK_loadPose =
(
	thePath = GetDir #scripts;
    filename = getOpenFileName caption:"Load POSE" filename:(thePath+"\\") types:"PoseTXT (*.txt)|*.txt|All (*.*)|*.*";
	if (filename == undefined) then		-- did the user cancel?
		return;

	format "-- loadPOSE: %\n" filename;

	r = loadPose filename;
	if (r == true) then
   		format "-- POSE % Loaded.\n" (getFileNameFile(filename));
)

/*
 * aK_resetPoseBlend() - Resets the pose blending
 */
fn aK_resetPoseBlend =
(
	aK_poseCntrl = #();		-- The control+value string that will set the pose
	aK_poseStart = #();		-- The initial value
	aK_poseEnd = #();		-- the end loaded value
	aK_poseType= #();		-- Type of value
	gc();	-- call Max's gc to cleanup garbage
)

/*
 * aK_addPoseBlend() - Adds an item to a pose blend.  This passes in the cntrl+val
 *		string, and the "saved state" itself to load.  This function then stores that as well
 *		as the "current state" so that we can blend between the two.
 *
 *		cntrl and val and type should be passed in as "strings"
 */
fn aK_addPoseBlend cntrl val type =
(
    cnt = aK_poseCntrl.count;
	cnt += 1;	-- lets add it!

	aK_poseCntrl[cnt] = (cntrl as string);
	aK_poseEnd[cnt] = (val as string);
    aK_poseType[cnt] = (type as string);

    if (type == "EulerAngles") then
		curVal = executeString("coordsys parent( ("+cntrl+" as EulerAngles) );");		-- get cur value
    else if (type == "Quat") then
		curVal = executeString("coordsys parent( ("+cntrl+" as Quat) );");		-- get cur value
	else
		curVal = executeString("coordsys parent( "+cntrl+" );");		-- get cur value
	aK_poseStart[cnt] = (curVal as string);

)


/*
 * aK_doPoseBlend() - This will look at the state of the slider and for each object
 *		in memory as a pose, will blend the values between the two.  It then also
 *		update the UI by calling the regular update proc...so that anything not
 *		caught normally is refreshed.
 */
fn aK_doPoseBlend =
(
    cnt = aK_poseCntrl.count;
	blend = attrRoll.slPose.value;

    if (cnt <= 0) then
	    return();

	disableSceneRedraw();

    for i in 1 to cnt do
		(
		str = stringstream "";

		if (aK_poseType[i] == "EulerAngles") then
			(
			appendString str ("eeS = "+aK_poseStart[i]+";\n");
			appendString str ("eeE = "+aK_poseEnd[i]+";\n");
			appendString str ("eeR = (eulerangles 0 0 0);\n");
			appendString str ("eeR.x = ( "+((1 - blend) as string)+"*(eeS.x)) + ("+(blend as string)+"*(eeE.x));\n");
			appendString str ("eeR.y = ( "+((1 - blend) as string)+"*(eeS.y)) + ("+(blend as string)+"*(eeE.y));\n");
			appendString str ("eeR.z = ( "+((1 - blend) as string)+"*(eeS.z)) + ("+(blend as string)+"*(eeE.z));\n");
			appendString str ("coordsys parent( "+aK_poseCntrl[i]+" = eeR );\n");
			executeString str;
			)
		else if (aK_poseType[i] == "Quat") then
			(
			appendString str ("qS = "+aK_poseStart[i]+";\n");
			appendString str ("qE = "+aK_poseEnd[i]+";\n");
			appendString str ("qR = (quat 0 0 0 1);\n");
			appendString str ("qR = slerp qS qE ("+(blend as string)+");\n");
			appendString str ("coordsys parent( "+aK_poseCntrl[i]+" = qR );\n");
			executeString str;
			)
/*
		else if (aK_poseType[i] == "boolean") then
			(
			appendString str ("bS = ("+aK_poseStart[i]+" as float);\n");
			appendString str ("bE = ("+aK_poseEnd[i]+" as float);\n");
			appendString str ("bR = 0.0;\n");
			appendString str ("bR = ( "+((1 - blend) as string)+"*(bS)) + ("+(blend as string)+"*(bE));\n");
			appendString str ("coordsys parent( "+aK_poseCntrl[i]+" = (bR as boolean) );\n");
			executeString str;
			)
*/
		else (
				appendString str ("coordsys parent( "+aK_poseCntrl[i]+" = ("+((1 - blend) as string)+"*("+aK_poseStart[i]+")) + ("+(blend as string)+"*("+aK_poseEnd[i]+")) );");
				executeString str;
				break;
			)
		)

    enableSceneRedraw();
    redrawViews();

    format "-- Pose Blended at: %\%.\n" (100.0*blend);

	update_attr_ui();	-- refresh as needed
)


/*
 * savePose() - This will save all DATA that we have for the SELECTED objects.
 *				to the given filename
 */
fn savePose filename =
(
    if (filename == undefined or filename == "") then
		(
		format "savePose: Invalid or missing filename.\n";
		return false;
		)

		-- open file
	f = createFile filename;
	if (f == undefined) then
		(
		format "savePose: Error creating output file: \"%\"\n" filename;
		return false;
		)

	setWaitCursor();

	format "-- attrKey: savePose DATA\n" to:f;
	format "\n" to:f;
	format "aK_resetPoseBlend();\n" to:f;
	format "\n" to:f;


		-- For each object...
	for oi in 1 to aK_objs.count do
		(
		objname = aK_objs[oi];
		obj = executeString("$'"+objname+"';");
			-- Only save if selected
		if (isSelected obj) then
			(
				-- First save PRS data
			pos = executeString("coordsys parent($'"+objname+"'.position);");
			rot = executeString("coordsys parent($'"+objname+"'.rotation as Quat);");
			scl = executeString("coordsys parent($'"+objname+"'.scale);");

				-- output PRS...make sure POS is LAST so we don't have rotation issues!
				--	from rotating around a different position!
			format "-- PRS DATA for %\n" objname to:f;
			format "if (isSelected $'%') then\n" objname to:f;
			format "    (\n" to:f;
			format "    if (aKloadPRS == 1) then\n" to:f;
			format "    	(\n" to:f;
			format "    	aK_addPoseBlend \"$'%'.scale\" \"%\" \"Point3\";\n" objname scl to:f;
			format "    	aK_addPoseBlend \"$'%'.rotation\" \"%\" \"Quat\";\n" objname rot to:f;
			format "    	aK_addPoseBlend \"$'%'.position\" \"%\" \"Point3\";\n" objname pos to:f;
			format "    	)\n" to:f;
			format "\n" to:f;
			format "    if (aKloadAtt == 1) then\n" to:f;
			format "    	(\n" to:f;

				-- Now do CA's if they exist
			start = aK_start[oi];
			end   = aK_end[oi];
			if (start >= 1 and end >= 1) then
				(
				format "	    -- CA DATA for %\n" objname to:f;
					-- do each saved ca index for this object
				for cai in start to end do
					(
						-- I really don't need to do boolean sep since it works like a float apparently...
					if (aK_type[cai] == #boolean) then
						(
						val = executeString(aK_cntrl[cai]+".value;");		-- get value
						format "	    aK_addPoseBlend \"%.value\" \"%\" \"boolean\";\n" aK_cntrl[cai] val to:f;
						)
					else (
						val = executeString(aK_cntrl[cai]+".value;");		-- get value
						format "	    aK_addPoseBlend \"%.value\" \"%\" \"%\";\n" aK_cntrl[cai] val (aK_type[cai] as string) to:f;
						)

					)
				)

			format "    	)\n" to:f;

			format "    )\n\n" to:f;

			)  -- end of isSelected if
		) -- end of oi loop

		-- close file
	close f;

    setArrowCursor();

    return true;

)

/*
 * loadPose() - This does the real loading of a POSE.  All it has to
 *				do is "execute" the script since the output format
 *				for saving was MAXscript commands.
 */
fn loadPose filename =
(
    if (filename == undefined or filename == "") then
		(
		format "loadPose: Invalid or missing filename.\n";
		return false;
		)

	f = openFile filename;
	if (f == undefined) then
		(
		format "loadPose: Error opening file for input: \"%\"\n" filename;
		return false;
		)

	setWaitCursor();

		-- Just execute all the data in it!
		-- This works because execute works on stringstreams!
	execute f;

		-- close file
	close f;

	setArrowCursor();

	attrRoll.slPose.enabled = true;
	attrRoll.slPose.value = 1.0;
    aK_doPoseBlend();  -- now call this to do the blending and effectively load the pos to start at 100%

    update_attr_ui();		-- Now make sure the keys on/off are correct.

    return true;
)


-- -------------------------- Keying Functions ---------------------
fn key_pos =
(
	setWaitCursor();
	for i in 1 to aK_objs.count do
	(
		str = "addnewkey ($'"+aK_objs[i]+"'.position.controller) currentTime;";
try (	executeString str;)
catch()
	)
	setArrowCursor();
)

fn key_pos_obj obj =
(
	setWaitCursor();
	str = "addnewkey ($'"+(obj.name)+"'.position.controller) currentTime;";
try (	executeString str; )
catch()

	setArrowCursor();
)



fn key_rot =
(
	setWaitCursor();
	for i in 1 to aK_objs.count do
	(
		str = "addnewkey ($'"+aK_objs[i]+"'.rotation.controller) currentTime;";
try (		executeString str;)
catch()
	)
	setArrowCursor();
)

fn key_rot_obj obj =
(
	setWaitCursor();
	str = "addnewkey ($'"+(obj.name)+"'.rotation.controller) currentTime;";
try (		executeString str;)
catch()
	setArrowCursor();
)

fn key_scl =
(
	setWaitCursor();
	for i in 1 to aK_objs.count do
	(
		str = "addnewkey ($'"+aK_objs[i]+"'.scale.controller) currentTime;";
try (		executeString str;)
catch()
	)
	setArrowCursor();
)

fn key_scl_obj obj =
(
	setWaitCursor();
	str = "addnewkey ($'"+(obj.name)+"'.scale.controller) currentTime;";
try (		executeString str;)
catch()
	setArrowCursor();
)


fn key_prs =
(
	setWaitCursor();
	disableSceneRedraw();

	key_pos();
	key_rot();
	key_scl();

	enableSceneRedraw();
	setArrowCursor();
)

fn zero_prs_obj obj =
(
	setWaitCursor();
--	disableSceneRedraw();

    str = stringstream "";
	appendString str ("coordsys parent($'"+(obj.name)+"'.scale = [1,1,1]);");
	appendString str ("coordsys parent($'"+(obj.name)+"'.rotation = (eulerangles 0 0 0));");
	appendString str ("coordsys parent($'"+(obj.name)+"'.position = [0,0,0]);");
try (	executeString str; )
catch()

--	enableSceneRedraw();
	setArrowCursor();
)


fn zero_prs =
(
	for i in 1 to aK_objs.count do
		(
		executeString("zero_prs_obj $'"+aK_objs[i]+"';");
		)
)



-- Keys the given attribute by index number from global array
fn key_attr i =
(
    if (i <= 0 or i > aK_attr.count) then
		return false;

	if (aK_type[i] == #boolean) then
	(
		same = executeString ("attrRoll."+aK_attrUI[i]+".state == "+aK_attr[i]+";");
		if (same) then
			return false;
	)

	str = "addnewkey ("+aK_cntrl[i]+") currentTime;";
try (
		k = executeString str;
		if (aK_type[i] == #boolean) then
		(
			-- update ui in case things have flopped for on/off
			executeString ("attrRoll."+aK_attrUI[i]+".state = "+aK_attr[i]+";");
		)
	)
catch ()

    update_attr_ui();		-- Now make sure the keys on/off are correct.

	return true;

)



-- ZERO's the given attribute by index number from global array
fn zero_attr i =
(
    if (i <= 0 or i > aK_attr.count) then
		return false;

	if (aK_type[i] == #boolean) then
	(
		same = executeString ("0 == "+aK_attr[i]+";");
		if (same) then
			return false;
	)

	if (aK_type[i] == #color) then
		(
        str = aK_attr[i]+" = (color 0 0 0);";
	    str += "attrRoll."+aK_attrUI[i]+".color = "+aK_attr[i]+";";
		)
	else
        str = aK_cntrl[i]+".value = 0;";
try (
		k = executeString str;
		if (aK_type[i] == #boolean) then
		(
			-- update ui in case things have flopped for on/off
			executeString ("attrRoll."+aK_attrUI[i]+".state = "+aK_attr[i]+";");
		)
	)
catch ()

    update_attr_ui();		-- Now make sure the keys on/off are correct.

	return true;

)

fn key_attrs =
(
	setWaitCursor();
	disableSceneRedraw();

	for i in 1 to aK_cntrl.count do
	(
		key_attr(i);
	)

	enableSceneRedraw();
	setArrowCursor();
)


fn zero_attrs =
(
	setWaitCursor();
	disableSceneRedraw();

	for i in 1 to aK_cntrl.count do
	(
		zero_attr(i);
	)

	enableSceneRedraw();
	redrawViews();
	setArrowCursor();
)


fn key_attrs_obj idx =
(
	setWaitCursor();
	disableSceneRedraw();

	start = aK_start[idx];
	end   = aK_end[idx];

	if (start < 1 or end < 1) then (
		enableSceneRedraw();
		setArrowCursor();
		return();
		)

	for i in start to end do
	(
		key_attr(i);
	)


	enableSceneRedraw();
	setArrowCursor();
)


fn zero_attrs_obj idx =
(
	setWaitCursor();
	disableSceneRedraw();

	start = aK_start[idx];
	end   = aK_end[idx];

	if (start < 1 or end < 1) then (
		enableSceneRedraw();
		setArrowCursor();
		return();
		)

	for i in start to end do
	(
		zero_attr(i);
	)


	enableSceneRedraw();
    redrawViews();
	setArrowCursor();
)


fn key_all =
(
	setWaitCursor();
	disableSceneRedraw();
	key_prs();
	key_attrs();
	enableSceneRedraw();
	setArrowCursor();
)

fn key_all_obj obj idx =
(
	setWaitCursor();
	disableSceneRedraw();

	key_pos_obj(obj);
	key_rot_obj(obj);
	key_scl_obj(obj);
	key_attrs_obj(idx);

	enableSceneRedraw();
	setArrowCursor();
)

-- --------------------------------------------------------------------------

/*
 * isKeyedNow() - Given a controller like $foo.myAttr.controller  this will
 *		returns true if the object has a keyframe at the current time, false otherwise.
 *
 *	Used to turn on/off the keyBracket display for spinners...
 */
fn isKeyedNow cntrl =
(
    if (cntrl == "" or cntrl == undefined) then
        return false;

    num = numKeys cntrl;
    if (num <= 0) then
        return false;

    for i in 1 to num do
        (
		kT = getKeyTime cntrl i;
		if (kT == currentTime) then
		    return true;
		)

    return false;

)

-- ------------------------------ UI Functions ----------------------

-- convert all spaces to underscore in the given string.
fn remove_spaces str =
(

	return str;

    if (str == undefined) then
		return undefined;

	for i in 1 to str.count do
	(
		if (str[i] == " ") then
			str[i] = "_";
	)
	return str;
)

-- Given a CA def and the parameter name, return the uiItem range for that parameter
fn get_ca_range cat paramName =
(

    if classof cat != AttributeDef do
		return undefined;

		-- fudge hand added ones thru script lack range data!?!?!
    if (cat.defdata == undefined) then
		return undefined;


	local def = cat.defdata[2];
	paramName = paramName as name;

	for uiItem in def where (uiItem.name as name) == paramName do
	(
--		format "uiItem.name: %\n" uiItem.name;
		if uiItem.ui as name == #spinner or uiItem.ui as name == #slider do  (
--			format "returning range is: %\n" uiItem.range;
			return (uiItem.range as string);
			)
	)

	undefined
)


fn create_attr_ui =
(

    deleteAllChangeHandlers id:#attrRotChange;

		-- remove old rollout if it exists
	str = "removeRollout attrRoll attrKeyRF";
	executeString str;
	attrRoll = undefined;

	aK_objs = #();
	aK_attr = #();
	aK_cntrl = #();
	aK_attrUI = #();
	aK_type = #();
	aK_start = #();
	aK_end = #();

	objN = 0;
	attrN = 0;

    sel = getCurrentSelection();	-- what objects have we?

	setWaitCursor();

	str = stringstream ""
	appendString str ("rollout attrRoll \"Attributes\"\n");
	appendString str ("(\n");
	appendString str ("\n");

-- format "DEBUG sel: %, %\n" sel.count sel;
	if (sel.count > 0) then
	(
		appendString str ("group \"Save/Load POSE\"");
		appendString str ("  (\n");
		appendString str ("button copyPOSE \"Copy\" width:35 height:18 offset:[2,0] tooltip:\"Copy POSE\" across:4;\n")
		appendString str ("button pastePOSE \"Paste\" width:35 height:18 offset:[4,0] tooltip:\"Paste POSE\"\n")
		appendString str ("button loadPOSE \"Load\" width:35 height:18 offset:[4,0] tooltip:\"Load POSE\"\n")
		appendString str ("button savePOSE \"Save\" width:35 height:18 offset:[5,0] tooltip:\"Save POSE\"\n")
		appendString str ("checkbutton cbLoadPRS \"PRS\" offset:[-7,-2] width:35 height:18 align:#center enabled:true tooltip:\"Pose Loading/Pasting will load PRS data.\" checked:"+((aKloadPRS == 1) as string)+" across:3;\n");
		appendString str ("checkbutton cbLoadAtt \"Attr\" offset:[-64,18] width:35 height:18 align:#center enabled:true tooltip:\"Pose Loading/Pasting will load Attr data.\" checked:"+((aKloadAtt == 1) as string)+";\n");
		appendString str ("slider slPose \"\" width:132 height:18 offset:[-72,0] range:[0,1,1] type:#float ticks:4 tooltip:\"Allows blending of the loaded/pasted pose.\" enabled:true;\n");
		appendString str ("  )\n");
		appendString str ("group \"Keying ALL Objects\"");
		appendString str ("  (\n");
		appendString str ("button keyPos \"POS\" width:40 height:18 offset:[-5,0] tooltip:\"Key Position for All Objs\" across:4;\n")
		appendString str ("button keyRot \"ROT\" width:40 height:18 offset:[-2,0] tooltip:\"Key Rotation for All Objs\";\n")
		appendString str ("button keyScl \"SCL\" width:40 height:18 offset:[3,0] tooltip:\"Key Scale for All Objs\";\n")
		appendString str ("button keyPRS \"PRS\" width:40 height:18 offset:[5,0] tooltip:\"Key Position, Rotation and Scale for All Objs\";\n")
		appendString str ("button keyAttrs \"Attr\" width:40 height:18 offset:[-5,0] tooltip:\"Key Attributes\" across:4;\n")
		appendString str ("button keyAll \"ALL\" width:40 height:18 offset:[-2,0] tooltip:\"Key Position, Rotation and Scale and Attributes for All Objs\";\n")
		appendString str ("button zeroAttrs \"Z Attr\" width:40 height:18 offset:[4,0] tooltip:\"Zero's All Attributes\";\n")
		appendString str ("button zeroAll \"Z ALL\" width:40 height:18 offset:[8,0] tooltip:\"Zero's All PRS and Attributes\";\n")
		appendString str ("  )\n");
	)

	for obj in sel do
	(
		objN = objN + 1;
		aK_objs[objN] = obj.name;
		aK_start[objN] = aK_end[objN] = 0;		-- set start/end array index of attrs for this obj to be nothing right now...

		appendString str ("group \""+(obj.name)+"\";";)
		appendString str (" (\n";)

		appendString str ("  button sel"+(objN as string)+" \"Select\" width:45 height:18 offset:[-6,0] align:#left tooltip:\"Select "+obj.name+"\" across:8;\n";)
		appendString str ("  button keyPos"+(objN as string)+" \"P\" width:15 height:18 offset:[24,0] tooltip:\"Key Position for "+obj.name+"\" ;\n";)
		appendString str ("  button keyRot"+(objN as string)+" \"R\" width:15 height:18 offset:[22,0] tooltip:\"Key Rotation for "+obj.name+"\" ;\n";)
		appendString str ("  button keyScl"+(objN as string)+" \"S\" width:15 height:18 offset:[18,0] tooltip:\"Key Scale for "+obj.name+"\" ;\n";)
		appendString str ("  button keyAttrs"+(objN as string)+" \"At\" width:18 height:18 offset:[18,0] tooltip:\"Key Attrs for "+obj.name+"\" ;\n";)
		appendString str ("  button keyAll"+(objN as string)+" \"All\" width:20 height:18 offset:[18,0] tooltip:\"Key PRS and Attrs for "+obj.name+"\" ;\n";)
		appendString str ("  button zeroAttrs"+(objN as string)+" \"z\" width:15 height:18 offset:[18,0] tooltip:\"Zero's All Attrs for "+obj.name+"\" ;\n";)
		appendString str ("  button zeroAll"+(objN as string)+" \"Z\" width:15 height:18 offset:[18,0] tooltip:\"Zero's PRS and All Attrs for "+obj.name+"\" ;\n";)
		appendString str ("\n");

		cntrl = "$'"+obj.name+"'.position.controller";
		ctype = executeString(cntrl+";");
		if (ctype as string == "Controller:Position_XYZ" and aKShowPRS == 1) then
		    (
			appendString str ("label pl"+(objN as string)+" \"P\" offset:[-20,2] across:4;");

			cntrl = "$'"+obj.name+"'.position.x_position.controller";
	        val = (executeString(cntrl+".value"));
			appendString str ("spinner px"+(objN as string)+" \"\" type:#float width:52 align:#left offset:[-30,2] range:[-9999999,9999999,"+(val as string)+"] controller:"+cntrl+";\n");
			cntrl = "$'"+obj.name+"'.position.y_position.controller";
	        val = (executeString(cntrl+".value"));
			appendString str ("spinner py"+(objN as string)+" \"\" type:#float width:52 align:#left offset:[-20,2] range:[-9999999,9999999,"+(val as string)+"] controller:"+cntrl+";\n");
			cntrl = "$'"+obj.name+"'.position.z_position.controller";
	        val = (executeString(cntrl+".value"));
			appendString str ("spinner pz"+(objN as string)+" \"\" type:#float width:52 align:#left offset:[-10,2] range:[-9999999,9999999,"+(val as string)+"] controller:"+cntrl+";\n");
		    appendString str ("\n");
			)

		cntrl = "$'"+obj.name+"'.rotation.controller";
		ctype = executeString(cntrl+";");
		if (ctype as string == "Controller:Euler_XYZ" and aKShowPRS == 1) then
		    (
			appendString str ("label rl"+(objN as string)+" \"R\" offset:[-20,2] across:4;");

			cntrl = "$'"+obj.name+"'.rotation.x_rotation.controller";
	        val = (executeString(cntrl+".value"));
			appendString str ("spinner rx"+(objN as string)+" \"\" type:#float width:52 align:#left offset:[-30,2] range:[-9999999,9999999,"+(val as string)+"] across:3;\n");

			cntrl = "$'"+obj.name+"'.rotation.y_rotation.controller";
	        val = (executeString(cntrl+".value"));
			appendString str ("spinner ry"+(objN as string)+" \"\" type:#float width:52 align:#left offset:[-20,2] range:[-9999999,9999999,"+(val as string)+"] ;\n");

			cntrl = "$'"+obj.name+"'.rotation.z_rotation.controller";
	        val = (executeString(cntrl+".value"));
			appendString str ("spinner rz"+(objN as string)+" \"\" type:#float width:52 align:#left offset:[-10,2] range:[-9999999,9999999,"+(val as string)+"] ;\n");
		    appendString str ("\n");

			)

		cntrl = "$'"+obj.name+"'.scale.controller";
		ctype = executeString(cntrl+";");
		if (ctype as string == "Controller:ScaleXYZ" and aKShowPRS == 1) then
		    (
			appendString str ("label sl"+(objN as string)+" \"S\" offset:[-20,2] across:4;");

			cntrl = "$'"+obj.name+"'.scale.x_scale.controller";
	        val = (executeString(cntrl+".value"));
			appendString str ("spinner sx"+(objN as string)+" \"\" type:#float width:52 align:#left offset:[-30,2] range:[-9999999,9999999,"+(val as string)+"] across:3;\n");

			cntrl = "$'"+obj.name+"'.scale.y_scale.controller";
	        val = (executeString(cntrl+".value"));
			appendString str ("spinner sy"+(objN as string)+" \"\" type:#float width:52 align:#left offset:[-20,2] range:[-9999999,9999999,"+(val as string)+"] ;\n");

			cntrl = "$'"+obj.name+"'.scale.z_scale.controller";
	        val = (executeString(cntrl+".value"));
			appendString str ("spinner sz"+(objN as string)+" \"\" type:#float width:52 align:#left offset:[-10,2] range:[-9999999,9999999,"+(val as string)+"] ;\n");
		    appendString str ("\n");

			)


		omod = obj.modifiers;		-- be sure to check modifiers too!
		omat = obj.material;
		cAnum=0;
		cAd = undefined;
		total = omod.count;
		if (omat != undefined) then
			total = total +1;


      for cAnum in 0 to total do
	  (
		local prefix= "";

		if (cAnum == 0) then
		(
		    cAD = custAttributes.getDefs obj;	-- to start get modifiers on base object
			prefix = ("$'"+obj.name+"'");

				-- Before looking for CA's...do special case MANIPULATOR grabbing of values or angle based on manip type.
			if (classOf obj == sliderManipulator) then
				(
				name = "value";
				uname = obj.sldName;
				attrN = attrN + 1;
				cntrl = prefix+"."+name+".controller";
				cn = executeString cntrl;
				-- if it's undefined, it means it hasn't been animated,
				-- and so doesn't have a controller assigned yet, so make one.
				if (cn == undefined) then
				(
					estr = cntrl+" = bezier_float();";
					executeString estr;
				)

				val	= (executeString(prefix+"."+name+";"));

				if (aKuseSliders == 0) then
					(
					appendString str ("label aKlbl"+name+(attrN as string)+" \""+uname+"\" align:#right offset:[20,3] across:3;\n");
					appendString str ("spinner aK"+name+(attrN as string)+" \"\" align:#right type:#float width:50 fieldWidth:50 offset:[28,2] scale:("+(obj.snapVal as string)+") range:["+(obj.minVal as string)+","+(obj.maxVal as string)+","+(val as string)+"] controller:("+cntrl+");\n");
					)
				else
					(
					appendString str ("label aKlbl"+name+(attrN as string)+" \""+uname+"\" align:#right offset:[-10,5] across:3;\n");
					appendString str ("slider aK"+name+(attrN as string)+" \"\" type:#float width:100 offset:[-10,-5] range:["+(obj.minVal as string)+","+(obj.maxVal as string)+","+(val as string)+"];\n");
					)
				appendString str ("button keyAttr"+(attrN as string)+" \"K\" align:#right toolTip:\"Key this attribute\" width:20;\n");
					-- store info
				aK_attr[attrN] = prefix+"."+name;
				aK_cntrl[attrN] = cntrl;
				aK_attrUI[attrN] = ("aK"+name+(attrN as string));
				aK_type[attrN] = #float;

				if (aK_start[objN] == 0) then
					aK_start[objN] = aK_end[objN] = attrN;		-- store start index of attrs for this obj
				else
					aK_end[objN] = attrN;					-- if we already store one, update the ending info
				)


		) else if (cAnum <= omod.count) then
		(
			local cmod = obj.modifiers[cAnum];	-- then start checking each modifier...
			try (
				cAD = custAttributes.getDefs cmod;
				)
			catch (
				cAD = undefined;
			)
			prefix ="$'"+obj.name+"'.'"+cmod.name+"'";
		) else
		(
			cAD = custAttributes.getDefs omat;	-- and finally check material if there is one
			prefix = "$'"+obj.name+"'.material";
		)


		if (cAD != undefined) then
		(
			for objDef in cAD do
			(
				pbArray = custAttributes.getPBlockDefs objdef;
				for a = 1 to pbArray.count do
				(
					itms = pbArray[a];

						-- Now go through attribute parameter names/info
					for p = 5 to itms.Count do
					(
						local name = "";
						local type = "";

							-- First item in the array is the name of the attr
--						format "\nDEBUG \t\t\t#name = %\n" itms[p][1]
						name = (itms[p][1] as string);
						rname = "'"+name+"'";
						uname = space2Under name;

--						format "DEBUG NAME %\n" name;
							-- Second item is another array, of info...
							-- Such as data type, default value, etc...
						for z = 1 to itms[p][2].Count by 2 do
						(
--							format "DEBUG \t\t\t% = %\n" itms[p][2][z] itms[p][2][z+1]
							if (itms[p][2][z] == #type) then
							(
								type = (itms[p][2][z+1]);
--								format "DEBUG TYPE %\n" type;
							)
						)

							-- At this point, we have an attr name,
							-- and the type of attribute it is.

						if (type == #float) then
						(
							attrN = attrN + 1;
--							cntrl = prefix+".Custom_Attributes."+name+".controller";
							cntrl = prefix+"."+rname+".controller";
							cn = executeString cntrl;
							-- if it's undefined, it means it hasn't been animated,
							-- and so doesn't have a controller assigned yet, so make one.
							if (cn == undefined) then
							(
								estr = cntrl+" = bezier_float();";
								executeString estr;
							)
--							format "DEBUG cntrl=%\n" cntrl;
							val	= (executeString(prefix+"."+rname+";"));

								-- figure range, which returns a string, then use filterString to get each start and end range value
							range = get_ca_range cAD[1] rname;


							if (range == undefined) then
								range = "[-9999,9999,0.0]";
							rangestr = filterString range "[],"


							if (aKuseSliders == 0) then
								(
								appendString str ("label aKlbl"+uname+(attrN as string)+" \""+name+"\" align:#right offset:[20,3] across:3;\n");
								appendString str ("spinner aK"+uname+(attrN as string)+" \"\" type:#float align:#right width:50 fieldWidth:50 offset:[28,2] range:["+(rangestr[1])+","+(rangestr[2])+","+(val as string)+"] controller:"+cntrl+";\n");
								)
							else
								(
								appendString str ("label aKlbl"+uname+(attrN as string)+" \""+name+"\" align:#right offset:[-10,5] across:3;\n");
								appendString str ("slider aK"+uname+(attrN as string)+" \"\" type:#float width:100 offset:[-10,-5] range:["+(rangestr[1])+","+(rangestr[2])+","+(val as string)+"];\n");
								)
							appendString str ("button keyAttr"+(attrN as string)+" \"K\" align:#right toolTip:\"Key this attribute\" width:20;\n");


								-- store info
							aK_attr[attrN] = prefix+"."+rname;
							aK_cntrl[attrN] = cntrl;
							aK_attrUI[attrN] = ("aK"+uname+(attrN as string));
							aK_type[attrN] = type;

							if (aK_start[objN] == 0) then
								aK_start[objN] = aK_end[objN] = attrN;		-- store start index of attrs for this obj
							else
								aK_end[objN] = attrN;					-- if we already store one, update the ending info
						)
						else if (type == #integer) then
						(
							attrN = attrN + 1;
--							cntrl = prefix+".Custom_Attributes."+name+".controller";
							cntrl = prefix+"."+rname+".controller";
							cn = executeString cntrl;
							-- if it's undefined, it means it hasn't been animated,
							-- and so doesn't have a controller assigned yet, so make one.
							if (cn == undefined) then
							(
								estr = cntrl+" = bezier_float();";
								executeString estr;
							)
--							format "DEBUG cntrl=%\n" cntrl;
							val	= (executeString(prefix+"."+rname+";"));

								-- figure range, which returns a string, then use filterString to get each start and end range value
							rname = "'"+name+"'";
							range = get_ca_range cAD[1] rname;
							if (range == undefined) then
								range = "[-9999,9999,0]";
							rangestr = filterString range "[],"

							if (aKuseSliders == 0) then
								(
								appendString str ("label aKlbl"+uname+(attrN as string)+" \""+name+"\" align:#right offset:[20,3] across:3;\n");
								appendString str ("spinner aK"+uname+(attrN as string)+" \"\" type:#integer width:50 fieldWidth:50 align:#right offset:[28,2] range:["+(rangestr[1])+","+(rangestr[2])+","+(val as string)+"] controller:"+cntrl+" across:2;\n");
								)
							else
								(
								appendString str ("label aKlbl"+uname+(attrN as string)+" \""+name+"\" align:#right offset:[-10,5] across:3;\n");
								appendString str ("slider aK"+uname+(attrN as string)+" \"\" type:#integer width:100 offset:[-10,-5] range:["+(rangestr[1])+","+(rangestr[2])+","+(val as string)+"];\n");
								)
							appendString str ("button keyAttr"+(attrN as string)+" \"K\" align:#right toolTip:\"Key this attribute\" width:20;\n");
								-- store info
							aK_attr[attrN] = prefix+"."+rname;
							aK_cntrl[attrN] = cntrl;
							aK_attrUI[attrN] = ("aK"+uname+(attrN as string));
							aK_type[attrN] = type;

							if (aK_start[objN] == 0) then
								aK_start[objN] = aK_end[objN] = attrN;		-- store start index of attrs for this obj
							else
								aK_end[objN] = attrN;					-- if we already store one, update the ending info
						)
						else if (type == #boolean) then
						(
							attrN = attrN + 1;
--							cntrl = prefix+".Custom_Attributes."+name+".controller";
							cntrl = prefix+"."+rname+".controller";
							cn = executeString cntrl;
							-- if it's undefined, it means it hasn't been animated,
							-- and so doesn't have a controller assigned yet, so make one.
							if (cn == undefined) then
							(
								estr = cntrl+" = on_off();";
								executeString estr;
							)
--							format "DEBUG cntrl=%\n" cntrl;
							if (executeString(prefix+"."+rname+";")) then
								val = "true";
							else
								val = "false";
							appendString str ("checkbox aK"+uname+(attrN as string)+" \""+name+"\" width:120 checked:"+val+" align:#left across:2;\n");
							appendString str ("button keyAttr"+(attrN as string)+" \"K\" align:#right toolTip:\"Key this attribute\" width:20;\n");
								-- store info
							aK_attr[attrN] = prefix+"."+rname;
							aK_cntrl[attrN] = cntrl;
							aK_attrUI[attrN] = ("aK"+uname+(attrN as string));
							aK_type[attrN] = type;

							if (aK_start[objN] == 0) then
								aK_start[objN] = aK_end[objN] = attrN;		-- store start index of attrs for this obj
							else
								aK_end[objN] = attrN;					-- if we already store one, update the ending info
						)
						else if (type == #color) then
						(
							attrN = attrN + 1;
--							cntrl = prefix+".Custom_Attributes."+name+".controller";
							cntrl = prefix+"."+rname+".controller";
							cn = executeString cntrl;
							-- if it's undefined, it means it hasn't been animated,
							-- and so doesn't have a controller assigned yet, so make one.
							if (cn == undefined) then
							(
								estr = cntrl+" = bezier_color();";
								executeString estr;
							)
--							format "DEBUG cntrl=%\n" cntrl;
							val	= (executeString(prefix+"."+rname+";"));
							appendString str ("colorpicker aK"+uname+(attrN as string)+" \""+name+"\" width:120 color:"+(val as string)+" align:#left across:2;\n");
							appendString str ("button keyAttr"+(attrN as string)+" \"K\" align:#right toolTip:\"Key this attribute\" width:20;\n");
								-- store info
							aK_attr[attrN] = prefix+"."+rname;
							aK_cntrl[attrN] = cntrl;
							aK_attrUI[attrN] = ("aK"+uname+(attrN as string));
							aK_type[attrN] = type;

							if (aK_start[objN] == 0) then
								aK_start[objN] = aK_end[objN] = attrN;		-- store start index of attrs for this obj
							else
								aK_end[objN] = attrN;					-- if we already store one, update the ending info
					)



					)
				)  -- end pbArray
			)  -- end objDef
		) -- end if cAD
      ) -- end of cAnum loop

		appendString str (" )\n");

	) -- end of obj loop



	appendString str ("on copyPOSE pressed do\n");
	appendString str ("(\n");
	appendString str ("     undo \"Copy POSE\" on ( aK_copyPose() )");
	appendString str (")\n");
	appendString str ("on pastePOSE pressed do\n");
	appendString str ("(\n");
	appendString str ("     undo \"Paste POSE\" on ( aK_pastePose() )");
	appendString str (")\n");
	appendString str ("on savePOSE pressed do\n");
	appendString str ("(\n");
	appendString str ("     undo \"Save POSE\" on ( aK_savePose() )");
	appendString str (")\n");
	appendString str ("on loadPOSE pressed do\n");
	appendString str ("(\n");
	appendString str ("     undo \"Load POSE\" on ( aK_loadPose() )");
	appendString str (")\n");
	appendString str ("on cbLoadPRS changed state do\n");
	appendString str ("(\n");
	appendString str ("    if (state) then\n");
	appendString str ("        aKloadPRS = 1;\n");
	appendString str ("    else\n");
	appendString str ("        aKloadPRS = 0;\n");
	appendString str (")\n");
	appendString str ("on cbLoadAtt changed state do\n");
	appendString str ("(\n");
	appendString str ("    if (state) then\n");
	appendString str ("        aKloadAtt = 1;\n");
	appendString str ("    else\n");
	appendString str ("        aKloadAtt = 0;\n");
	appendString str (")\n");
	appendString str ("on slPose changed val do\n");
	appendString str ("(\n");
	appendString str ("        aK_doPoseBlend();\n");
	appendString str (")\n");


	appendString str ("on keyPos pressed do\n");
	appendString str ("(\n");
	appendString str ("     undo \"attrKey: Key Position\" on ( key_pos() )");
	appendString str (")\n");
	appendString str ("on keyRot pressed do\n");
	appendString str ("(\n");
	appendString str ("     undo \"attrKey: Key Rotation\" on ( key_rot() )");
	appendString str (")\n");
	appendString str ("on keyScl pressed do\n");
	appendString str ("(\n");
	appendString str ("     undo \"attrKey: Key Scale\" on ( key_scl() )");
	appendString str (")\n");
	appendString str ("on keyPRS pressed do\n");
	appendString str ("(\n");
	appendString str ("     undo \"attrKey: Key PRS\" on ( key_prs() )");
	appendString str (")\n");
	appendString str ("on keyAttrs pressed do\n");
	appendString str ("(\n");
	appendString str ("     undo \"attrKey: Key Attributes\" on ( key_attrs() )");
	appendString str (")\n");
	appendString str ("on keyAll pressed do\n");
	appendString str ("(\n");
	appendString str ("     undo \"attrKey: Key PRS And Attributes\" on ( key_all() )");
	appendString str (")\n");
	appendString str ("on zeroAttrs pressed do\n");
	appendString str ("(\n");
	appendString str ("	undo \"attrKey: ZERO Attributes\" on ( zero_attrs(); )\n");
	appendString str (")\n");
	appendString str ("on zeroAll pressed do\n");
	appendString str ("(\n");
	appendString str ("	undo \"attrKey: ZERO PRS and Attributes\" on ( zero_prs(); zero_attrs(); )\n");
	appendString str (")\n");

	for i in 1 to objN do
	(

		appendString str ("on sel"+(i as string)+" pressed do\n");
		appendString str ("(\n");
		appendString str ("	select($"+ak_Objs[i]+");\n");
		appendString str (")\n");

		appendString str ("on keyPos"+(i as string)+" pressed do\n");
		appendString str ("(\n");
		appendString str ("	undo \"attrKey: Key OBJ Position\" on ( key_pos_obj($'"+ak_Objs[i]+"'); )\n");
		appendString str (")\n");

		appendString str ("on keyRot"+(i as string)+" pressed do\n");
		appendString str ("(\n");
		appendString str ("	undo \"attrKey: Key OBJ Rotation\" on ( key_rot_obj($'"+ak_Objs[i]+"'); )\n");
		appendString str (")\n");

		appendString str ("on keyScl"+(i as string)+" pressed do\n");
		appendString str ("(\n");
		appendString str ("	undo \"attrKey: Key OBJ Scale\" on ( key_scl_obj($'"+ak_Objs[i]+"'); )\n");
		appendString str (")\n");

		appendString str ("on keyAttrs"+(i as string)+" pressed do\n");
		appendString str ("(\n");
		appendString str ("	undo \"attrKey: Key OBJ Attributes\" on ( key_attrs_obj("+(i as string)+") )\n");
		appendString str (")\n");

		appendString str ("on keyAll"+(i as string)+" pressed do\n");
		appendString str ("(\n");
		appendString str ("	undo \"attrKey: Key OBJ PRS and Attributes\" on ( key_all_obj $'"+(ak_Objs[i])+"' "+(i as string)+")\n");
		appendString str (")\n");

		appendString str ("on zeroAttrs"+(i as string)+" pressed do\n");
		appendString str ("(\n");
		appendString str ("	undo \"attrKey: ZERO OBJ Attributes\" on ( zero_attrs_obj("+(i as string)+") )\n");
		appendString str (")\n");

		appendString str ("on zeroAll"+(i as string)+" pressed do\n");
		appendString str ("(\n");
		appendString str ("	undo \"attrKey: ZERO OBJ PRS and Attributes\" on ( zero_prs_obj($'"+ak_Objs[i]+"'); zero_attrs_obj("+(i as string)+") )\n");
		appendString str (")\n");


		cntrl = "$'"+ak_Objs[i]+"'.rotation.controller";
		ctype = executeString(cntrl+";");
		if (ctype as string == "Controller:Euler_XYZ" and aKShowPRS == 1) then
		    (
			appendString str ("on rx"+(i as string)+" changed val do\n");
			appendString str ("(\n");
			appendString str ("    undo \"attrKey: Rotate Change\" on (\n");
			appendString str ("    	$'"+ak_Objs[i]+"'.rotation.x_rotation.controller.value = (val);\n");
			appendString str ("		);");
			appendString str (")\n");
		    appendString str ("\n");

			appendString str ("on ry"+(i as string)+" changed val do\n");
			appendString str ("(\n");
			appendString str ("    undo \"attrKey: Rotate Change\" on (\n");
			appendString str ("    	$'"+ak_Objs[i]+"'.rotation.y_rotation.controller.value = (val);\n");
			appendString str ("		);");
			appendString str (")\n");
		    appendString str ("\n");

			appendString str ("on rz"+(i as string)+" changed val do\n");
			appendString str ("(\n");
			appendString str ("    undo \"attrKey: Rotate Change\" on (\n");
			appendString str ("    	$'"+ak_Objs[i]+"'.rotation.z_rotation.controller.value = (val);\n");
			appendString str ("		);");
			appendString str (")\n");
		    appendString str ("\n");

			)

		cntrl = "$'"+ak_Objs[i]+"'.scale.controller";
		ctype = executeString(cntrl+";");
		if (ctype as string == "Controller:ScaleXYZ" and aKShowPRS == 1) then
		    (
			appendString str ("on sx"+(i as string)+" changed val do\n");
			appendString str ("(\n");
			appendString str ("    undo \"attrKey: Scale Change\" on (\n");
			appendString str ("    	$'"+ak_Objs[i]+"'.scale.x_scale.controller.value = (val);\n");
			appendString str ("		);");
			appendString str (")\n");
		    appendString str ("\n");

			appendString str ("on sy"+(i as string)+" changed val do\n");
			appendString str ("(\n");
			appendString str ("    undo \"attrKey: Scale Change\" on (\n");
			appendString str ("    	$'"+ak_Objs[i]+"'.scale.y_scale.controller.value = (val);\n");
			appendString str ("		);");
			appendString str (")\n");
		    appendString str ("\n");

			appendString str ("on sz"+(i as string)+" changed val do\n");
			appendString str ("(\n");
			appendString str ("    undo \"attrKey: Scale Change\" on (\n");
			appendString str ("    	$'"+ak_Objs[i]+"'.scale.z_scale.controller.value = (val);\n");
			appendString str ("		);");
			appendString str (")\n");
		    appendString str ("\n");
			)

	)

	for i in 1 to aK_attr.count do
	(
		appendString str ("on keyAttr"+(i as string)+" pressed do\n");
		appendString str ("(\n");
		appendString str ("    undo \"attrKey: Key Attribute\" on ( key_attr("+(i as string)+") );");
		appendString str (")\n");

			-- need to do this for these types, since there is no direct controller link up
		if (aK_type[i] == #boolean) then
		(
			appendString str ("on "+aK_attrUI[i]+" changed state do\n");
			appendString str ("(\n");
			appendString str ("        "+aK_attr[i]+" = state;");
			appendString str (")\n");

		)
		else if (aK_type[i] == #color) then
		(
			appendString str ("on "+aK_attrUI[i]+" changed newCol do\n");
			appendString str ("(\n");
			appendString str ("        "+aK_attr[i]+" = newCol;");
			appendString str (")\n");
		)
		else if (aKuseSliders == 1) then
		(
			if (aK_type[i] == #float or aK_type[i] == #integer) then
			(
				appendString str ("on "+aK_attrUI[i]+" changed val do\n");
				appendString str ("(\n");
				appendString str ("        "+aK_attr[i]+" = val;");
				appendString str (")\n");
			)
		)

	)

	appendString str ("\n");
	appendString str (")\n\n");
	executeString str;			-- now make the attrRoll


--	format "DEBUG str:%\n" str;

	str = "addRollout attrRoll attrKeyRF;";
	executeString str;


    str = stringstream "\n";
	for i in 1 to objN do
		(
		cntrl = "$'"+ak_Objs[i]+"'.rotation.controller";
		ctype = executeString(cntrl+";");
		cntrl2 = "$'"+ak_Objs[i]+"'.scale.controller";
		ctype2 = executeString(cntrl2+";");
		if ( ((ctype as string == "Controller:Euler_XYZ") or (ctype2 as string == "Controller:ScaleXYZ")) and aKShowPRS == 1) then
		    (
			appendString str ("when transform $'"+ak_Objs[i]+"' changes id:#attrRotChange do\n");
			appendString str ("(\n");
		    if (ctype as string == "Controller:Euler_XYZ") then
				(
				appendString str ("    attrRoll.rz"+(i as string)+".value = $'"+ak_Objs[i]+"'.rotation.z_rotation.controller.value;\n");
				appendString str ("    attrRoll.ry"+(i as string)+".value = $'"+ak_Objs[i]+"'.rotation.y_rotation.controller.value;\n");
				appendString str ("    attrRoll.rx"+(i as string)+".value = $'"+ak_Objs[i]+"'.rotation.x_rotation.controller.value;\n");
				)
		    if (ctype2 as string == "Controller:ScaleXYZ") then
				(
				appendString str ("    attrRoll.sz"+(i as string)+".value = $'"+ak_Objs[i]+"'.scale.z_scale.controller.value;\n");
				appendString str ("    attrRoll.sy"+(i as string)+".value = $'"+ak_Objs[i]+"'.scale.y_scale.controller.value;\n");
				appendString str ("    attrRoll.sx"+(i as string)+".value = $'"+ak_Objs[i]+"'.scale.x_scale.controller.value;\n");
				)
			appendString str (") -- end\n");
		    appendString str ("\n");
			)
		)

    executeString str;

	setArrowCursor();

    update_attr_ui();		-- Now make sure the keys on/off are correct.

)


-- Called when time changes
-- Refreshed needed UI items for those that don't have controller tie-ins
fn update_attr_ui =
(
    for i in 1 to aK_objs.count do
		(
		cntrl = "$'"+ak_Objs[i]+"'.rotation.controller";
		ctype = executeString(cntrl+";");
		if (ctype as string == "Controller:Euler_XYZ" and aKShowPRS == 1) then
		    (
		    local str = stringstream "";
			appendString str ("attrRoll.rz"+(i as string)+".value = $'"+ak_Objs[i]+"'.rotation.z_rotation.controller.value;\n");
			appendString str ("attrRoll.ry"+(i as string)+".value = $'"+ak_Objs[i]+"'.rotation.y_rotation.controller.value;\n");
			appendString str ("attrRoll.rx"+(i as string)+".value = $'"+ak_Objs[i]+"'.rotation.x_rotation.controller.value;\n");
		    appendString str ("\n");
			executeString str;
			)

		cntrl2 = "$'"+ak_Objs[i]+"'.scale.controller";
		ctype2 = executeString(cntrl2+";");
		if (ctype2 as string == "Controller:ScaleXYZ" and aKShowPRS == 1) then
		    (
		    local str = stringstream "";

			appendString str ("attrRoll.sz"+(i as string)+".value = $'"+ak_Objs[i]+"'.scale.z_scale.controller.value;\n");
			appendString str ("attrRoll.sy"+(i as string)+".value = $'"+ak_Objs[i]+"'.scale.y_scale.controller.value;\n");
			appendString str ("attrRoll.sx"+(i as string)+".value = $'"+ak_Objs[i]+"'.scale.x_scale.controller.value;\n");
		    appendString str ("\n");
			executeString str;
			)
		)

	for i in 1 to aK_attr.count do
	(
			-- need to do this for these types, since there is no direct controller link up
		if (aK_type[i] == #boolean) then
		(
try(
			local str = "attrRoll."+aK_attrUI[i]+".state = "+aK_attr[i]+";";
			executeString str;
)
catch ()
		)
		else if (aK_type[i] == #color) then
		(
			local str = "attrRoll."+aK_attrUI[i]+".color = "+aK_attr[i]+";";
			executeString str;
		)
		else if (aKuseSliders == 1) then
		(
			if (aK_type[i] == #float or aK_type[i] == #integer) then
			(
			local str = "attrRoll."+aK_attrUI[i]+".value = "+aK_attr[i]+";";
			executeString str;
			)
		)
	)
)

-- ----------------------- Call Back Wrapper Functions ----------
-- Called whenever time is changed
fn aK_timechanged =
(
--	format "-- time changed.\n";
--    aK_resetPoseBlend();	-- i guess we don't have to do this
	update_attr_ui();
)

-- Called whenever an object is selected/deselected
fn aK_selchanged =
(
--	format "-- sel changed.\n";
	create_attr_ui();
)

-- ------------------------------ UI Stuff --------------------------

rollout optionsRoll "Options"
(
	checkbutton refreshSelected "Selection Change Redraws UI" width:160 align:#center enabled:true tooltip:"Will cause UI to be updated when object(s) are selected/de-selected" checked:true;
	checkbutton showPRS "Display PRS Spinners" width:160 align:#center enabled:true tooltip:"Will show PRS values for Position_XYZ, Euler_XYZ and ScaleXYZ controlled objects." checked:true;
	label lblType "U.I. Type:" offset:[-12,0] across:2;
	radiobuttons slideType labels:#("Spinner", "Slider") default:1 columns:2 offset:[-5,0];

/*
	checkbutton watchTime "Redraw UI on Time Change" width:160 align:#center enabled:true tooltip:"All of UI items are refreshed when time changes.  Turn off for faster playback" checked:true;
*/

		-- This occurs once the floater window is created
	on optionsRoll open do
	(
		format "-- attrKey starting up.\n";
		attrKeyRunning = 1;
		aK_resetPoseBlend();
		create_attr_ui();

		registerTimeCallback aK_timechanged;
	    callbacks.addScript #selectionSetChanged "aK_selchanged()" id:#aK_selChangeCB;
	)

	on optionsRoll close do
	(
		format "-- attrKey shutting down.\n";
		attrKeyRunning = 0;
		unRegisterTimeCallback aK_timechanged;
	    callbacks.removeScripts id:#aK_selChangeCB;
	    deleteAllChangeHandlers id:#attrRotChange;
	)


    on showPRS changed state do
	(
	    if (state) then
			aKShowPRS = 1;
		else
			aKShowPRS = 0;
		create_attr_ui();	-- remake ui
	)

    on refreshSelected changed state do
	(
		if (state) then
		(
			create_attr_ui();
			callbacks.addScript #selectionSetChanged "aK_selchanged()" id:#aK_selChangeCB;
		) else
		(
	    	callbacks.removeScripts id:#aK_selChangeCB;
		)
	)

    on slideType changed state do
    (
		aKuseSliders = state - 1;	-- becomes 0 or 1
		create_attr_ui();	-- remake ui
	)


/*
    on watchTime changed state do
	(
		if (state) then
		(
			update_attr_ui();
			registerTimeCallback aK_timechanged;
		) else
		(
			unRegisterTimeCallback aK_timechanged;
		)
	)
*/

)




rollout aboutRoll "About attrKey"
(
	label abtlbl1 "";
	label abtlbl2 "by";
	label abtlbl3 "Michael B. Comet";
	label abtlbl4 "http://www.comet-cartoons.com/";
	label abtlbl5 "comet@comet-cartoons.com";
	label abtlbl6 "";
	label abtlbl7 "Copyright 2002 Michael B. Comet";
	label abtlbl8 "All Rights Reserved";
	label abtlbl9 "";

	on aboutroll open do
	(
		abtlbl1.text = aKtitlever;
	)
)

rollout attrRoll "Attributes"
(
)

-- Do any preRollout preprocessing here
-- create the rollout window and add the rollout
if (attrKeyRunning != 1) then
(
	attrKeyRF = newRolloutFloater ("Comet attrKey "+aKtitlever) 210 410;
	addRollout aboutRoll attrKeyRF rolledUp:true;
	addRollout optionsRoll attrKeyRF rolledUp:true;
	addRollout attrRoll attrKeyRF;
)


-- Do any postRollout preprocessing here


--end of attrKey.ms

